#include "BoostGraph.h"
#include "Helper.h"
#include <map>
#include <stack>
#include <fstream>
#include "boost/graph/dijkstra_shortest_paths.hpp"
#include <boost/graph/johnson_all_pairs_shortest.hpp>
#include "boost/graph/isomorphism.hpp"

//=========================================================================================================================

BoostGraph::BoostGraph()
{
}

//=========================================================================================================================

BoostGraph::~BoostGraph()
{
}

//=========================================================================================================================

bool BoostGraph::LoadVienna(ViennaFile& viennaFile, bool displayDebugText)
{
	char dotBracket;
	stack<string> pairs;
	map<string, VertexDescriptor> vertexMap;
	string fullVertexName = "";

	VertexDescriptor start = AddVertex("Start");
	VertexDescriptor previousPair = start;

	for(int i = 0; i < viennaFile.baseSequence.length(); ++i)
	{
		dotBracket = viennaFile.dotBracket[i];
		fullVertexName = viennaFile.baseSequence[i] + Helper::Int2String(i + 1);

		if(displayDebugText)
			cout << "	Base: " + fullVertexName + " '" + dotBracket + "'" << endl;

		if(dotBracket == '.')
		{
			vertexMap[fullVertexName] = AddVertex(fullVertexName);

			if(pairs.empty())
				add_edge(start, vertexMap[fullVertexName], uGraph);
			else
				add_edge(vertexMap[pairs.top()], vertexMap[fullVertexName], uGraph);
		}
		else if(dotBracket == '(')
		{
			vertexMap[fullVertexName] = AddVertex(fullVertexName);

			// Create edge to last pair
			if(pairs.empty())
				add_edge(previousPair, vertexMap[fullVertexName], uGraph);
			else
				add_edge(vertexMap[pairs.top()], vertexMap[fullVertexName], uGraph);

			pairs.push(fullVertexName);
			previousPair = vertexMap[fullVertexName];
		}
		else if(dotBracket == ')')
		{
			if(pairs.empty())
			{
				cerr << "	Error: Secondary interaction mis-match, premature ')' encountered" << endl;
				uGraph.clear();
				return false;
			}

			uGraph[vertexMap[pairs.top()]].id += (":" + fullVertexName);
			pairs.pop();
		}
	}

	if(!pairs.empty())
	{
		cerr << "	Error: Found " + Helper::Int2String(pairs.size()) + " unpaired interactions" << endl;
		uGraph.clear();
		return false;
	}

	cout << endl;

	sourceViennaFile = viennaFile;

	return true;
}

//=========================================================================================================================

bool BoostGraph::LoadSIF(vector<vector<string>>& data, bool displayDebugText)
{
	// Strategy:
	// 1. Load all nodes into a map<string, VertexDescriptor>
	// 2. Go through the data again and create edges for each node

	int verticesNotFound = 0;
	map<string, VertexDescriptor> vertexInfo;

	for(vector<vector<string>>::iterator dataIter = data.begin();
		dataIter != data.end();
		++dataIter)
	{
		if(vertexInfo.find((*dataIter)[0]) == vertexInfo.end()) // Vertex does not already exist
		{
			if(displayDebugText)
				cout << "Creating vertex: " << (*dataIter)[0] << endl;

			vertexInfo[(*dataIter)[0]] = add_vertex(uGraph);
			uGraph[vertexInfo[(*dataIter)[0]]].id = (*dataIter)[0];
		}
		else
		{
			if(displayDebugText)
				cout << "Vertex already exists: " << (*dataIter)[0] << endl;
		}
	}

	for(vector<vector<string>>::iterator dataIter = data.begin();
		dataIter != data.end();
		++dataIter)
	{
		if(dataIter->size() < 3) // If there are no nodes connected the current node, skip it
			continue;

		for(vector<string>::iterator tokenIter = (*dataIter).begin() + 2; // Skip node and relationship type
			tokenIter != (*dataIter).end();
			++tokenIter)
		{
			// 1. Check if node exists
			// 2. Output statement
			// 3. Call boost::add_edge(...)

			if(vertexInfo.find((*tokenIter)) == vertexInfo.end())
			{
				if(displayDebugText)
					cerr << "Vertex not found: " << (*tokenIter) << endl;
	
				++verticesNotFound;
			}
			else
			{
				if(!edge(vertexInfo[(*dataIter).front()], vertexInfo[(*tokenIter)], uGraph).second &&
				   !edge(vertexInfo[(*tokenIter)], vertexInfo[(*dataIter).front()], uGraph).second)
				{
					if(displayDebugText)
						cout << "Creating edge between " << (*dataIter)[0] << " and " << (*tokenIter) << endl;

					add_edge(vertexInfo[(*dataIter).front()], vertexInfo[(*tokenIter)], uGraph);
				}
				else
				{
					if(displayDebugText)
						cout << "Edge between " << (*dataIter)[0] << " and " << (*tokenIter) << " already exists!" << endl;
				}
			}
		}
	}

	cout << endl
	 	 << "Graph successfully constructed:"
		 << "==============================="
  		 << "Vertices not found  : " << verticesNotFound << endl
		 << "Vertices            : " << GetNoVertices() << endl
		 << "Edges               : " << GetNoEdges() << endl << endl;

	return true;
}

//=========================================================================================================================

bool BoostGraph::ExportToCytoscape(const string& filePath)
{
	ofstream file;
	file.open(filePath.c_str(), ios_base::out);

	if(!file.is_open())
	{
		cerr << "	Error: Could not open requested file for output" << endl;
		return false;
	}

	VertexIterator vertexIter;
	VertexIterator vertexIterEnd;
	AdjacencyIterator adjacencyIter;
	AdjacencyIterator adjacencyIterEnd;
	map<string, string> connectivityMap;

	// Print out all vertices and their connections
	for(tie(vertexIter, vertexIterEnd) = vertices(uGraph); vertexIter != vertexIterEnd; ++vertexIter)
	{
		if(uGraph[*vertexIter].id.find(':') != string::npos || uGraph[*vertexIter].id == "Start")
		{
			file << uGraph[*vertexIter].id << " interactsWith";
		}
		else
			continue;

		// Access all adjacent vertices of the current vertex and print them out
		for(tie(adjacencyIter, adjacencyIterEnd) = adjacent_vertices(*vertexIter, uGraph);
			adjacencyIter != adjacencyIterEnd; ++adjacencyIter)
		{
			if(connectivityMap.find(uGraph[*adjacencyIter].id) == connectivityMap.end())//connectivityMap.end())
			{
				// The adjacenct vertex does not connect with the current one (yet)
				file << " " << uGraph[*adjacencyIter].id;
				connectivityMap[uGraph[*vertexIter].id] = uGraph[*adjacencyIter].id;
			}
		}

		file << "\n";
	}

	file.close();

	return true;
}

//=========================================================================================================================

void BoostGraph::Print()
{
	VertexIterator vertexIter;
	VertexIterator vertexIterEnd;
	AdjacencyIterator adjacencyIter;
	AdjacencyIterator adjacencyIterEnd;

	// Print out all vertices and their connections
	for(tie(vertexIter, vertexIterEnd) = vertices(uGraph); vertexIter != vertexIterEnd; ++vertexIter)
	{
		cout << uGraph[*vertexIter].id << " -> ";

		// Access all adjacent vertices of the current vertex and print them out
		for(tie(adjacencyIter, adjacencyIterEnd) = adjacent_vertices(*vertexIter, uGraph);
			adjacencyIter != adjacencyIterEnd;)
		{
			cout << uGraph[*adjacencyIter].id;

			++adjacencyIter;

			if(adjacencyIter != adjacencyIterEnd)
				cout << ", ";
		}

		cout << endl;
	}
}

//=========================================================================================================================

void BoostGraph::PrintVertices()
{
	VertexIterator vertexIter;
	VertexIterator vertexIterEnd;

	for(tie(vertexIter, vertexIterEnd) = vertices(uGraph); vertexIter != vertexIterEnd;)
	{
		cout << uGraph[*vertexIter].id;

		++vertexIter;

		if(vertexIter != vertexIterEnd)
			cout << endl;
	}
}

//=========================================================================================================================

void BoostGraph::DijkstraAllPairsShortestPaths()
{
	VertexDescriptor source = vertex(0, uGraph);
	vector<VertexDescriptor> path(GetNoVertices());
	vector<int> distances(GetNoVertices());

	property_map<UndirectedGraph, int VertexProperty::*>::type indexMap = get(&VertexProperty::index, uGraph);
	property_map<UndirectedGraph, int EdgeProperty::*>::type weightMap = get(&EdgeProperty::weight, uGraph);

	//dijkstra_shortest_paths(uGraph, source, predecessor_map(&path[0]).distance_map(&distances[0]));

	//dijkstra_shortest_paths(uGraph,
	//						source,
	//						make_iterator_property_map(path.begin(),
	//					   							   get(&VertexProperty::index, uGraph)),
	//					    make_iterator_property_map(distances.begin(),
	//												   get(&VertexProperty::index, uGraph)),
	//						weightMap, indexMap,
	//						std::less<int>(),
	//						closed_plus<int>(),
	//						(std::numeric_limits<int>::max)(),
	//						0,
	//						default_dijkstra_visitor());

	for(int i = 0; i < distances.size(); ++i)
		cout << distances[i] << ", ";
}

//=========================================================================================================================

void BoostGraph::JohnsonsAllPairsShortestPaths()
{
	vector<int> distances(GetNoVertices(), 1000);
	cout << "Size: " << GetNoVertices();
	int distanceMatrix[5][5];

	property_map<UndirectedGraph, int VertexProperty::*>::type indexmap = get(&VertexProperty::index, uGraph);
	property_map<UndirectedGraph, int EdgeProperty::*>::type weightmap = get(&EdgeProperty::weight, uGraph);

	johnson_all_pairs_shortest_paths(uGraph, distanceMatrix, indexmap, weightmap, 0);

	for(int i = 0; i < 5; ++i)
	{
		for(int j = 0; j < 5; ++j)
		{
			cout << "D[" << i << "][" << j << "]: " << distanceMatrix[i][j] << endl;
		}
	}
}

//=========================================================================================================================

bool BoostGraph::Compare(BoostGraph& comparableGraph)
{
	vector<VertexDescriptor> f(GetNoVertices());
	property_map<UndirectedGraph, int VertexProperty::*>::type indexMap1 = get(&VertexProperty::index, uGraph);
	property_map<UndirectedGraph, int VertexProperty::*>::type indexMap2 = get(&VertexProperty::index, comparableGraph.GetGraph());

	bool isIsomorphic = isomorphism(uGraph, comparableGraph.GetGraph(), isomorphism_map(make_iterator_property_map(f.begin(), indexMap1, f[0])));

	cout << "Is isomorphic?: " << (isIsomorphic ? "True" : "False") << endl << endl
		 << "f:" << endl;

	for(int i = 0; i < f.size(); ++i)
		cout << f[i] << endl;

	//vector<VertexDescriptor> f(GetNoVertices());
	//property_map<UndirectedGraph, int VertexProperty::*>::type indexMap1 = get(&VertexProperty::index, uGraph);
	//property_map<UndirectedGraph, int VertexProperty::*>::type indexMap2 = get(&VertexProperty::index, comparableGraph.GetGraph());
	//
	//property_map<int, int>::type inDegreeMap;

	//VertexIterator vertexIter;
	//VertexIterator vertexIterEnd;

	//for(tie(vertexIter, vertexIterEnd) = vertices(uGraph); vertexIter != vertexIterEnd; ++vertexIter)
	//{
	//	put(inDegreeMap, *vertexIter, out_degree(*vertexIter, uGraph));
	//}

	//return isomorphism(uGraph, comparableGraph.GetGraph(),
	//				   isomorphism_map(make_iterator_property_map(f.begin(), indexMap1, f[0])),
	//				   degree_vertex_invariant<std::vector<int>, UndirectedGraph>(),
	//				   degree_vertex_invariant<std::vector<int>, UndirectedGraph>(),
	//				   vertex_invariant2.max(),
	//				   indexMap1,
	//				   indexMap2);

	return false;
}

//=========================================================================================================================

bool BoostGraph::AddVertex(string base, int position, bool debug)
{
	if(position < 0 || position > sourceViennaFile.baseSequence.length())
	{
		cerr << "	Position is out of range";
		return false;
	}

	uGraph.clear();

	sourceViennaFile.baseSequence = sourceViennaFile.baseSequence.substr(0, position) +
		base + sourceViennaFile.baseSequence.substr(position);

	sourceViennaFile.dotBracket = sourceViennaFile.dotBracket.substr(0, position) +
		"." + sourceViennaFile.dotBracket.substr(position);

	if(debug)
		cout << "	Result: " << sourceViennaFile.baseSequence << endl
			 << "		" << sourceViennaFile.dotBracket;

	if(LoadVienna(sourceViennaFile, false))
		return true;

	return false;
}

//=========================================================================================================================

VertexDescriptor BoostGraph::AddVertex(string id)
{
	VertexDescriptor vertexDescriptor = add_vertex(uGraph);
	uGraph[vertexDescriptor].id = id;

	return vertexDescriptor;
}

//=========================================================================================================================

bool BoostGraph::RemoveVertex(int position, bool debug)
{
	if(position > sourceViennaFile.baseSequence.length() || position < 0)
	{
		cerr << "	Position is out of range";
		return false;
	}

	if(sourceViennaFile.dotBracket.compare(position - 1, 1, ".") != 0)
	{
		cerr << "	Position does not contain a vertex";
		return false;
	}

	uGraph.clear();

	sourceViennaFile.baseSequence = sourceViennaFile.baseSequence.substr(0, position - 1) +
		sourceViennaFile.baseSequence.substr(position);

	sourceViennaFile.dotBracket = sourceViennaFile.dotBracket.substr(0, position - 1) +
		sourceViennaFile.dotBracket.substr(position);

	if(debug)
		cout << "	Result: " << sourceViennaFile.baseSequence << endl
			 << "		" << sourceViennaFile.dotBracket;

	if(LoadVienna(sourceViennaFile, false))
		return true;

	return false;
}

//=========================================================================================================================

bool BoostGraph::AddEdge(string base1, string base2, int position1, int position2, bool debug)
{
	if(position1 > sourceViennaFile.baseSequence.length() || position1 < 0
		|| position2 > sourceViennaFile.baseSequence.length() || position2 < 0)
	{
		cerr << "	One or more positions are out of range";
		return false;
	}

	uGraph.clear();

	sourceViennaFile.baseSequence = sourceViennaFile.baseSequence.substr(0, position1) +
		base1 + sourceViennaFile.baseSequence.substr(position1);

	sourceViennaFile.baseSequence = sourceViennaFile.baseSequence.substr(0, position2 + 1) +
		base2 + sourceViennaFile.baseSequence.substr(position2 + 1);

	sourceViennaFile.dotBracket = sourceViennaFile.dotBracket.substr(0, position1) +
		"(" + sourceViennaFile.dotBracket.substr(position1);

	sourceViennaFile.dotBracket = sourceViennaFile.dotBracket.substr(0, position2 + 1) +
		")" + sourceViennaFile.dotBracket.substr(position2 + 1);

	if(debug)
		cout << "	Result: " << sourceViennaFile.baseSequence << endl
			 << "		" << sourceViennaFile.dotBracket;

	if(LoadVienna(sourceViennaFile, false))
		return true;

	return false;
}

//=========================================================================================================================

bool BoostGraph::RemoveEdge(int position1, int position2, bool debug)
{
	if(position1 > sourceViennaFile.baseSequence.length() || position1 < 0
		|| position2 > sourceViennaFile.baseSequence.length() || position2 < 0)
	{
		cout << "	One or more positions are out of range";
		return false;
	}

	if(sourceViennaFile.dotBracket.compare(position1 - 1, 1, "(") != 0
		|| sourceViennaFile.dotBracket.compare(position2 - 1, 1, ")") != 0)
	{
		cout << "	One or more positions do not contain an edge";
		return false;
	}

	uGraph.clear();

	sourceViennaFile.baseSequence = sourceViennaFile.baseSequence.substr(0, position1 - 1) +
		sourceViennaFile.baseSequence.substr(position1);
	sourceViennaFile.baseSequence = sourceViennaFile.baseSequence.substr(0, position2 - 2) +
		sourceViennaFile.baseSequence.substr(position2 - 1);

	sourceViennaFile.dotBracket = sourceViennaFile.dotBracket.substr(0, position1 - 1) +
		sourceViennaFile.dotBracket.substr(position1);
	sourceViennaFile.dotBracket = sourceViennaFile.dotBracket.substr(0, position2 - 2) +
		sourceViennaFile.dotBracket.substr(position2 - 1);

	if(debug)
		cout << "	Result: " << sourceViennaFile.baseSequence << endl
			 << "		" << sourceViennaFile.dotBracket;

	if(LoadVienna(sourceViennaFile, false))
		return true;

	return false;
}

//=========================================================================================================================

bool BoostGraph::Change(int position, string base, bool debug)
{
	if(position > sourceViennaFile.baseSequence.length() || position < 0)
	{
		cerr << "	Position is out of range";
		return false;
	}

	sourceViennaFile.baseSequence.replace(position - 1, 1, base);

	if(debug)
		cout << "	Result: " << sourceViennaFile.baseSequence << endl
			 << "		" << sourceViennaFile.dotBracket;

	return true;
}

//=========================================================================================================================

int BoostGraph::GetNoVertices()
{
	return num_vertices(uGraph);
}

//=========================================================================================================================

int BoostGraph::GetNoEdges()
{
	return num_edges(uGraph);
}

//=========================================================================================================================

int BoostGraph::GetNetworkDiameter()
{
	return 0;
}

//=========================================================================================================================

bool BoostGraph::IsEmpty()
{
	return (GetNoVertices() == 0); // If there are no vertices there can be no edges
}

//=========================================================================================================================

UndirectedGraph& BoostGraph::GetGraph()
{
	return uGraph;
}

//=========================================================================================================================